'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function ({ config = {}, plugins = {} } = {}) {
  const stack = [];

  config = Object.assign({
    baseUrl: '/'
  }, config);

  function normalizeError(err) {
    if (err && (0, _isFunction2.default)(err.toJSON)) {
      return err.toJSON();
    }

    return {
      message: err.message
    };
  }

  const dispatcher = async (req, res) => {
    try {
      req.config = config;
      for (let call of stack) {
        await call(req, res);
      }
    } catch (err) {
      if (err.nonce) {
        return;
      }
      debug(err);
      res.status(err.status || 500).json(normalizeError(err));
    }
  };

  function registerPlugin(plugin) {
    if ((0, _isFunction2.default)(plugin)) {
      return (...spec) => {
        stack.push(plugin(...spec));
        return dispatcher;
      };
    }

    if ((0, _isObject2.default)(plugin)) {
      const container = {};
      (0, _forEach2.default)(plugin, (name, key) => {
        container[key] = (...spec) => {
          stack.push(name(...spec));
          return dispatcher;
        };
      });

      return container;
    }

    throw new Error(`The provided plugin has invalid type "${typeof plugin}"`);
  }

  // plugins
  function plugin(name, plugin) {
    if (dispatcher.hasOwnProperty(name)) {
      throw new Error(`The plugin with name ${name} is already registered or would overwrite a builtin method.`);
    }
    Object.defineProperty(dispatcher, name, {
      value: registerPlugin(plugin)
    });
    return dispatcher;
  }

  Object.defineProperty(dispatcher, 'plugin', {
    value: plugin
  });

  (0, _forEach2.default)(plugins, (fn, name) => plugin(name, fn));

  return dispatcher;
};

var _debug = require('debug');

var _debug2 = _interopRequireDefault(_debug);

var _isFunction = require('lodash/isFunction');

var _isFunction2 = _interopRequireDefault(_isFunction);

var _isObject = require('lodash/isObject');

var _isObject2 = _interopRequireDefault(_isObject);

var _forEach = require('lodash/forEach');

var _forEach2 = _interopRequireDefault(_forEach);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const debug = (0, _debug2.default)('netiam:dispatcher');